# Author: Sébastien Verel
# email : verel@univ-littoral.fr
# Date  : 02/22/2017
# Example of fitness landscape analysis
# The 51st CREST Open Workshop - Tutorial on Landscape Analysis
#
# Squares Problem (SP):
#   One solution: 
#      set of n squares of size c
#   Fitness function:
#      Number of brown points in the squares - number of blue points in the squares
#
# Part 1: length of adaptive walks to compare diffculty of optimization problems


# Include ggplot2 for 'beautiful' plots
library(ggplot2)

# number of squares
n <- 5  
# size of the squares
c <- 200 

# Drawing functions: draw all points of the Squares Problem
draw <- function(pts) {
  p <- ggplot(pts, aes(x=x1, y=x2, color=factor(coul))) + theme_bw() + geom_point(size=1) + xlim(0, 1000) + ylim(0, 1000) + theme(legend.position="none") + coord_fixed(ratio=1) + scale_color_manual(values=c("#56B4E9", "#E69F00"))
  print(p)
  return(p)
}

# Drawing functions: add on a graphics the squares of the Squares Problem 
draw_solution <- function(p, x, color = "red") {
  df <- data.frame(x1=x[,1],x2=x[,2])
  p <- p + geom_rect(data = df, aes(xmin=x1, xmax=x1+c, ymin=x2, ymax=x2+c), colour= color, fill = NA)
  print(p)
  
  return(p)
}

# Fitness function

# helper function
in.squares <- function(x) {
  return(function(pt) {
    j <- 1
    while(j <= nrow(x) && !(x[j,1] <= pt[1] && pt[1] <= x[j,1] + c && x[j,2] <= pt[2] && pt[2] <= x[j,2] + c)) {
      j = j + 1
    }
    
    if (j <= nrow(x))
      return(pt[3])
    else
      return(0)
  })
}

# Create the fitness function of a SP defined by points
fitness_create <- function(points) {
  # function which associates the number of brown points - number of blue points
  f <- function(x) {
    in.square <- in.squares(x)
    r <- apply(points, 1, in.square)
    return(sum(r))
  }
  
  return(f)
}

pb1_create <- function() {
  set.seed(1234)
  n.blue <- 400
  n.red  <- 150
  x1.blue <- ceiling(rnorm(n = n.blue) * 250 + 500)
  x2.blue <- ceiling(rnorm(n = n.blue) * 250 + 500)
  x1.red <- ceiling(rnorm(n = n.red) * 100 + 250)
  x2.red <- ceiling(rnorm(n = n.red) * 100 + 250)
  points <- data.frame(x1 = c(x1.blue, x1.red), x2 = c(x2.blue, x2.red), coul = c(rep(1, n.blue),rep(-1, n.red)))
  #write.table(points, file ="data/pb_1.csv", quote = FALSE, row.names = FALSE)
  return(points)
} 

pb2_create <- function() {
  set.seed(1234)
  n.blue <- 400
  n.red  <- 150
  x1.blue <- ceiling(rnorm(n = n.blue) * 250 + 500)
  x2.blue <- ceiling(rnorm(n = n.blue) * 250 + 500)
  x1.red <- ceiling(rnorm(n = n.red) * 50 + 450)
  x2.red <- ceiling(rnorm(n = n.red) * 50 + 450)
  points <- data.frame(x1 = c(x1.blue, x1.red), x2 = c(x2.blue, x2.red), coul = c(rep(1, n.blue),rep(-1, n.red)))
  #write.table(points, file ="data/pb_2.csv", quote = FALSE, row.names = FALSE)
  return(points)
} 

# Random initialization of a solution
# random coordinates for n squares between (0,0) and (1000-c, 1000-c)
init <- function() {
  return(matrix(ceiling(runif(n=2 * n, min = 0, max = 1000 - c)), ncol = 2, byrow = TRUE))  
}

# hill-climbing local search based on a neighborhood
hc_ngh <- function(fitness, neighborhood, x) {
  # number of neighbors
  size = nrow(neighborhood)
  
  # randomized order
  ids = sample(1:size) 
  i = 1
  
  # fitness value of the current solution x
  fitx = fitness(x)
  
  step    = 0  # number of step of the adaptive walk
  neval = 0  # number of evaluation
  while (i <= size) {
    # one random neighbor y
    delta <- neighborhood[ids[i],]
    y <- x
    y[delta[1],] <- y[delta[1],] + c(delta[2], delta[3])
    fity <- fitness(y)
    neval <- neval + 1 
    if (fitx < fity) {
      # accept the neighbor y
      x <- y
      fitx = fity
      step <- step + 1
      #      cat(step, i, fit, "\n")
      # start again with another randomized order
      ids <- sample(1:size) 
      i <- 1
    } else {
      # next neighbor
      i <- i + 1  
    }
  }
  
  # return the local optimum with the statistics
  return(list(x = x, fitness = fitx, step = step, neval = neval))
}

# create a neighborhood 
# Each neighbor move (x,y) of one square in the square [(-r, -r), (r, r)]
neighborhood_create <- function(r) {
  m <- matrix(ncol = 3, nrow = 0)
  for(i in 1:n) {
    for(d1 in seq(-r, r, by = 5)) {
      for(d2 in seq(-r, r, by = 5)) {
        if (d1 != 0 && d2 != 0) {
          m <- rbind(m, c(i, d1, d2))
        }
      }
    }
  }
  return(m)
}

main <- function() {
  # create points from random or from a file
  points1 <- pb1_create()
  #points1 <- read.table("data/pb_1.csv", sep = " ", header = TRUE)

  # draw the points of the SP
  p <- draw(points1) 
  
  # create a random solution and draw it
  set.seed(12345)
  x <- init()
  p <- draw_solution(p, x, "green")
  
  # fitness function
  ff1 <- fitness_create(points1)
  cat(ff1(x), "\n")
  
  # neighborhood
  neighborhood <- neighborhood_create(25)
  
  # run the hill-climber
  result <- hc_ngh(ff1, neighborhood, x)
  
  # draw the result
  p <- draw_solution(p, result$x, "red")
  cat(result$fitness, "\n")
}

adaptive_length <- function(fitness, neighborhood, nb) {
  # data frame for the results
  results <- data.frame(step=integer(), neval=integer(), fitness=integer())
  
  for(i in 1:nb) {
    # initial random solution
    x <- init()
    # run hc
    result <- hc_ngh(fitness, neighborhood, x)
    # print and save the result
    cat("step:", result$step, "neval:", result$neval, "fitness:", result$fitness, "x:", result$x, "\n")
    results <- rbind(results, data.frame(step=result$step, neval=result$neval, fitness=result$fitness))
  }
  
  return(results)
}


# the goal is to compare the adaptive length of two SP 
main_adaptive_length_analysis <- function() {
  # create points from random or from a file SP1
  points1 <- pb1_create()
  #points1 <- read.table("data/pb_1.csv", sep = " ", header = TRUE)
  p <- draw(points1) 
  ff1 <- fitness_create(points1)
  
  # create points from random or from a file SP2
  points2 <- pb2_create() 
  #points2 <- read.table("data/pb_2.csv", sep = " ", header = TRUE)
  p <- draw(points2) 
  ff2 <- fitness_create(points2)
  
  # neighborhood
  neighborhood <- neighborhood_create(25)

  # run the adaptive walks  
  set.seed(23456)
  al1 <- adaptive_length(bff1, neighborhood, 30)
  al2 <- adaptive_length(bff2, neighborhood, 30)
  # cheating mode: read the results from files
  #al1 <- read.table("data/adaptive_length_pb_1.csv", sep = " ", header = TRUE)
  #al2 <- read.table("data/adaptive_length_pb_2.csv", sep = " ", header = TRUE)
  
  # analyze the results
  boxplot(list(pb1 = al1$step, pb2 = al2$step))
  summary(al1)
  summary(al2)
  wilcox.test(al1$step, al2$step)
} 

