# Author: Sébastien Verel
# email : verel@univ-littoral.fr
# Date  : 02/22/2017
# Example of fitness landscape analysis
# The 51st CREST Open Workshop - Tutorial on Landscape Analysis
#
# Squares Problem (SP):
#   One solution: 
#      set of n squares of size c
#   Fitness function:
#      Number of brown points in the squares - number of blue points in the squares
#
# Part 2: ruggedness estimated by random walks


# Include ggplot2 for 'beautiful' plots
library(ggplot2)

# number of squares
n <- 5  
# size of the squares
c <- 200 

# Drawing functions: draw all points of the Squares Problem
draw <- function(pts) {
  p <- ggplot(pts, aes(x=x1, y=x2, color=factor(coul))) + theme_bw() + geom_point(size=1) + xlim(0, 1000) + ylim(0, 1000) + theme(legend.position="none") + coord_fixed(ratio=1) + scale_color_manual(values=c("#56B4E9", "#E69F00"))
  print(p)
  return(p)
}

# Drawing functions: add on a graphics the squares of the Squares Problem 
draw_solution <- function(p, x, color = "red") {
  df <- data.frame(x1=x[,1],x2=x[,2])
  p <- p + geom_rect(data = df, aes(xmin=x1, xmax=x1+c, ymin=x2, ymax=x2+c), colour= color, fill = NA)
  print(p)
  
  return(p)
}

# Fitness function

# helper function
in.squares <- function(x) {
  return(function(pt) {
    j <- 1
    while(j <= nrow(x) && !(x[j,1] <= pt[1] && pt[1] <= x[j,1] + c && x[j,2] <= pt[2] && pt[2] <= x[j,2] + c)) {
      j = j + 1
    }
    
    if (j <= nrow(x))
      return(pt[3])
    else
      return(0)
  })
}

# Create the fitness function of a SP defined by points
fitness_create <- function(points) {
  # function which associates the number of brown points - number of blue points
  f <- function(x) {
    in.square <- in.squares(x)
    r <- apply(points, 1, in.square)
    return(sum(r))
  }
  
  return(f)
}

pb1_create <- function() {
  set.seed(1234)
  n.blue <- 400
  n.red  <- 150
  x1.blue <- ceiling(rnorm(n = n.blue) * 250 + 500)
  x2.blue <- ceiling(rnorm(n = n.blue) * 250 + 500)
  x1.red <- ceiling(rnorm(n = n.red) * 100 + 250)
  x2.red <- ceiling(rnorm(n = n.red) * 100 + 250)
  points <- data.frame(x1 = c(x1.blue, x1.red), x2 = c(x2.blue, x2.red), coul = c(rep(1, n.blue),rep(-1, n.red)))
  #write.table(points, file ="data/pb_1.csv", quote = FALSE, row.names = FALSE)
  return(points)
} 

pb2_create <- function() {
  set.seed(1234)
  n.blue <- 400
  n.red  <- 150
  x1.blue <- ceiling(rnorm(n = n.blue) * 250 + 500)
  x2.blue <- ceiling(rnorm(n = n.blue) * 250 + 500)
  x1.red <- ceiling(rnorm(n = n.red) * 50 + 450)
  x2.red <- ceiling(rnorm(n = n.red) * 50 + 450)
  points <- data.frame(x1 = c(x1.blue, x1.red), x2 = c(x2.blue, x2.red), coul = c(rep(1, n.blue),rep(-1, n.red)))
  #write.table(points, file ="data/pb_2.csv", quote = FALSE, row.names = FALSE)
  return(points)
} 

# Random initialization of a solution
# random coordinates for n squares between (0,0) and (1000-c, 1000-c)
init <- function() {
  return(matrix(ceiling(runif(n=2 * n, min = 0, max = 1000 - c)), ncol = 2, byrow = TRUE))  
}

# create a mutation operator:
# p: mutation rate of each square
# r: range of coordinates mutation
mutation_create <- function(p, r) {
  values <- c(-r:r)
  mut <- function(x) {
    y <- x
    for(i in 1:nrow(x)) {
      if (runif(n=1) < p) {
        delta <- sample(values, size = 2, replace = TRUE)
        while (delta[1] == 0 && delta[2] == 0)
          delta <- sample(values, size = 2, replace = TRUE)
        y[i,] <- y[i,] + delta
      }
    }
    return(y)
  }
  
  return(mut)
}

# Perform a random walk and return fitness values
random_walk <- function(fitness, mutation, wlen) {
  x <- init()
  # vector of fitness values of the random walk
  fit_values <- c(fitness(x))
  
  for(i in 1:(wlen-1)) {
    x <- mutation(x)
    fit_values <- c(fit_values, fitness(x))
  }
  
  return(fit_values)
}


main <- function() {
  # create points from random or from a file
  points1 <- pb1_create()
  #points1 <- read.table("data/pb_1.csv", sep = " ", header = TRUE)

  # draw the points of the SP
  p <- draw(points1) 

  # fitness function
  ff1 <- fitness_create(points1)

  # mutation
  mutation <- mutation_create(p=1/n, r=40)
  
  # run a random walk (to do)
  set.seed(12345)
  fitness_values1 <- random_walk(ff1, mutation, 1000)

  # plot the fitness values of the walk
  plot(fitness_values1, type = "l")
  # autocorrelation function
  a1 <- acf(fitness_values1, lag.max = 100)
  # values of the first autocorrelation coefficient
  cat(a1$acf[2], "\n")
  
  # with SP 2
  points2 <- pb2_create()
  #points2 <- read.table("data/pb_2.csv", sep = " ", header = TRUE)
  ff2 <- fitness_create(points2)

  fitness_values2 <- random_walk(ff2, mutation, 1000)

  # plot the fitness values of the walks
  plot(fitness_values1, type = "l", col = "blue")
  lines(fitness_values2, col = "red")

  # autocorrelation function
  a2 <- acf(fitness_values2, lag.max = 100)
  # values of the first autocorrelation coefficient
  cat(a2$acf[2], "\n")
  
}

