# Author: Sébastien Verel
# email : verel@univ-littoral.fr
# Date  : 02/22/2017
# Example of fitness landscape analysis
# The 51st CREST Open Workshop - Tutorial on Landscape Analysis
#
# Squares Problem (SP):
#   One solution: 
#      set of n squares of size c
#   Fitness function:
#      Number of brown points in the squares - number of blue points in the squares
#
# Part 4: performance explaination based on simple fitness landscape features 


# Include ggplot2 for 'beautiful' plots
library(ggplot2)
library(GGally)

# number of squares
n <- 5  
# size of the squares
c <- 200 

# Drawing functions: draw all points of the Squares Problem
draw <- function(pts) {
  p <- ggplot(pts, aes(x=x1, y=x2, color=factor(coul))) + theme_bw() + geom_point(size=1) + xlim(0, 1000) + ylim(0, 1000) + theme(legend.position="none") + coord_fixed(ratio=1) + scale_color_manual(values=c("#56B4E9", "#E69F00"))
  print(p)
  return(p)
}

# Drawing functions: add on a graphics the squares of the Squares Problem 
draw_solution <- function(p, x, color = "red") {
  df <- data.frame(x1=x[,1],x2=x[,2])
  p <- p + geom_rect(data = df, aes(xmin=x1, xmax=x1+c, ymin=x2, ymax=x2+c), colour= color, fill = NA)
  print(p)
  
  return(p)
}

# Fitness function

# helper function
in.squares <- function(x) {
  return(function(pt) {
    j <- 1
    while(j <= nrow(x) && !(x[j,1] <= pt[1] && pt[1] <= x[j,1] + c && x[j,2] <= pt[2] && pt[2] <= x[j,2] + c)) {
      j = j + 1
    }
    
    if (j <= nrow(x))
      return(pt[3])
    else
      return(0)
  })
}

# Create the fitness function of a SP defined by points
fitness_create <- function(points) {
  # function which associates the number of brown points - number of blue points
  f <- function(x) {
    in.square <- in.squares(x)
    r <- apply(points, 1, in.square)
    return(sum(r))
  }
  
  return(f)
}

pb1_create <- function() {
  set.seed(1234)
  n.blue <- 400
  n.red  <- 150
  x1.blue <- ceiling(rnorm(n = n.blue) * 250 + 500)
  x2.blue <- ceiling(rnorm(n = n.blue) * 250 + 500)
  x1.red <- ceiling(rnorm(n = n.red) * 100 + 250)
  x2.red <- ceiling(rnorm(n = n.red) * 100 + 250)
  points <- data.frame(x1 = c(x1.blue, x1.red), x2 = c(x2.blue, x2.red), coul = c(rep(1, n.blue),rep(-1, n.red)))
  #write.table(points, file ="data/pb_1.csv", quote = FALSE, row.names = FALSE)
  return(points)
} 

pb2_create <- function() {
  set.seed(1234)
  n.blue <- 400
  n.red  <- 150
  x1.blue <- ceiling(rnorm(n = n.blue) * 250 + 500)
  x2.blue <- ceiling(rnorm(n = n.blue) * 250 + 500)
  x1.red <- ceiling(rnorm(n = n.red) * 50 + 450)
  x2.red <- ceiling(rnorm(n = n.red) * 50 + 450)
  points <- data.frame(x1 = c(x1.blue, x1.red), x2 = c(x2.blue, x2.red), coul = c(rep(1, n.blue),rep(-1, n.red)))
  #write.table(points, file ="data/pb_2.csv", quote = FALSE, row.names = FALSE)
  return(points)
} 

# Random initialization of a solution
# random coordinates for n squares between (0,0) and (1000-c, 1000-c)
init <- function() {
  return(matrix(ceiling(runif(n=2 * n, min = 0, max = 1000 - c)), ncol = 2, byrow = TRUE))  
}

# create a mutation operator:
# p: mutation rate of each square
# r: range of coordinates mutation
mutation_create <- function(p, r) {
  values <- c(-r:r)
  mut <- function(x) {
    y <- x
    for(i in 1:nrow(x)) {
      if (runif(n=1) < p) {
        delta <- sample(values, size = 2, replace = TRUE)
        while (delta[1] == 0 && delta[2] == 0)
          delta <- sample(values, size = 2, replace = TRUE)
        y[i,] <- y[i,] + delta
      }
    }
    return(y)
  }
  
  return(mut)
}

# Perform a random walk and return fitness values
random_walk <- function(fitness, mutation, wlen) {
  x <- init()
  # vector of fitness values of the random walk
  fit_values <- c(fitness(x))
  
  for(i in 1:(wlen-1)) {
    x <- mutation(x)
    fit_values <- c(fit_values, fitness(x))
  }
  
  return(fit_values)
}

# Compute the neutral rate estimated by a random walk
neutral_rate <- function(fit_values) {
  l <- length(fit_values)
  
  nrate = sum(fit_values[1:l-1] == fit_values[2:l]) / (l-1)
  
  return(nrate)
}

# Compute main fitness landscape features based on a random random
fitness_landscape_features <- function(fit_values) {
  l <- length(fit_values)
  
  # autocorrelation function
  acf1 <- acf(fit_values, plot = FALSE, lag.max = 400)
  # autocorrelation length
  len1 <- which.max(acf1$acf < 4 / sqrt(l)) - 1
  if (len1 == 0) len1 = 400
  
  # remove neutral neighbors
  fval <- c(fit_values[0])
  for(i in 2:l) {
    if (fit_values[i] != fit_values[i-1]) {
      fval <- c(fval, fit_values[i])
    }
  } 
  # autocorrelation function without neutral neighbors
  acf2 <- acf(fval, plot = FALSE, lag.max = 300)
  # autocorrelation length
  len2 <- which.max(acf2$acf < 4 / sqrt(l)) - 1
  if (len2 == 0) len2 = 300
  
  # all statistics
  res <- list(avg = mean(fit_values), sd = sd(fit_values), rho1 = acf1$acf[2], k = len1, rho1u = acf2$acf[2], ku = len2, nrate = neutral_rate(fit_values))
  
  return(res)
}

# Perform a random walk (of length 1000) on each SP problem
random_walk_samplings <- function() {
  set.seed(12345)
  mutation <- mutation_create(p = 0.2, r = 40)
  
  for(i in 1:20) {
    cat("pb", i, "\n")
    # fitness function of the SP problem i
    points <- read.table(paste0("data/pb_", i,".csv"), sep = " ", header = TRUE)
    ff <- fitness_create(points)
    
    # random walk
    results <- random_walk(ff, mutation, 1000)
    
    # save into file
    write.table(results, file = paste0("data/rw_02_40_", i, ".csv"), quote = FALSE, row.names = FALSE, col.names = FALSE)
  }
}
  
fitness_landscape_analysis <- function() {
  # analysis of each SP problem
  df <- data.frame()
  for(i in 1:20) {
    values = read.table(paste0("data/rw_02_40_", i,".csv"), sep = " ", header = FALSE)$V1
    res <- fitness_landscape_features(values)
    res$pb = i
    df <- rbind(df, res)
  }
    
  return(df)
}

# Bind the ILS performance to the data frame
ils_performance <- function(df) {
  perf <- c()
  for(i in 1:20) {
    values = read.table(paste0("data/perf_ils_", i,".csv"), sep = " ", header = TRUE)$fitness
    perf <- c(perf, mean(values))
  }
  df$ils = perf
  # normalized performance by the "random search"
  df$normperf = (df$ils - df$avg) / df$sd

  return(df)
}

main <- function() {
  # random walk on each problem
  random_walk_samplings()
  
  # fitness landscape analysis
  df <- fitness_landscape_analysis()
  
  # Bind the performance of ILS
  df <- ils_performance(df)
  
  # Analyze the results
  View(df)
}

