# Author: Sébastien Verel
# email : verel@univ-littoral.fr
# Date  : 02/22/2017
# Example of fitness landscape analysis
# The 51st CREST Open Workshop - Tutorial on Landscape Analysis
#
# Bi-objective Squares Problem (SP2):
#   One solution: 
#      set of n squares of size c
#   Fitness function:
#      (Number of brown points in the squares, number of blue points in the squares)
#
# Part 5: multi-objective fitness landscape features 


# Include ggplot2 for 'beautiful' plots
library(ggplot2)
library(GGally)

# number of squares
n <- 5  
# size of the squares
c <- 200 

# Drawing functions: draw all points of the Squares Problem
draw <- function(pts) {
  p <- ggplot(pts, aes(x=x1, y=x2, color=factor(coul))) + theme_bw() + geom_point(size=1) + xlim(0, 1000) + ylim(0, 1000) + theme(legend.position="none") + coord_fixed(ratio=1) + scale_color_manual(values=c("#56B4E9", "#E69F00"))
  print(p)
  return(p)
}

# Drawing functions: add on a graphics the squares of the Squares Problem 
draw_solution <- function(p, x, color = "red") {
  df <- data.frame(x1=x[,1],x2=x[,2])
  p <- p + geom_rect(data = df, aes(xmin=x1, xmax=x1+c, ymin=x2, ymax=x2+c), colour= color, fill = NA)
  print(p)
  
  return(p)
}

# Fitness function

# helper function
in.squares2 <- function(x) {
  return(function(pt) {
    j <- 1
    while(j <= nrow(x) && !(x[j,1] <= pt[1] && pt[1] <= x[j,1] + c && x[j,2] <= pt[2] && pt[2] <= x[j,2] + c)) {
      j = j + 1
    }
    
    if (j <= nrow(x)) {
      return(1)
    } else
      return(0)
  })
}

# Create the biobjective fitness function of a SP2 defined by points
fitness2_create <- function(points) {
  # function which associates the (number of brown points, number of blue points)
  pOne <- points[points$coul == 1,]
  pMinusOne <- points[points$coul == -1,]
  f <- function(x) {
    insquare2 <- in.squares2(x)
    rOne <- apply(pOne, 1, insquare2)
    rMinusOne <- apply(pMinusOne, 1, insquare2)
    return(c(sum(rOne), sum(rMinusOne)))
  }
  
  return(f)
}

pb1_create <- function() {
  set.seed(1234)
  n.blue <- 400
  n.red  <- 150
  x1.blue <- ceiling(rnorm(n = n.blue) * 250 + 500)
  x2.blue <- ceiling(rnorm(n = n.blue) * 250 + 500)
  x1.red <- ceiling(rnorm(n = n.red) * 100 + 250)
  x2.red <- ceiling(rnorm(n = n.red) * 100 + 250)
  points <- data.frame(x1 = c(x1.blue, x1.red), x2 = c(x2.blue, x2.red), coul = c(rep(1, n.blue),rep(-1, n.red)))
  #write.table(points, file ="data/pb_1.csv", quote = FALSE, row.names = FALSE)
  return(points)
} 

pb2_create <- function() {
  set.seed(1234)
  n.blue <- 400
  n.red  <- 150
  x1.blue <- ceiling(rnorm(n = n.blue) * 250 + 500)
  x2.blue <- ceiling(rnorm(n = n.blue) * 250 + 500)
  x1.red <- ceiling(rnorm(n = n.red) * 50 + 450)
  x2.red <- ceiling(rnorm(n = n.red) * 50 + 450)
  points <- data.frame(x1 = c(x1.blue, x1.red), x2 = c(x2.blue, x2.red), coul = c(rep(1, n.blue),rep(-1, n.red)))
  #write.table(points, file ="data/pb_2.csv", quote = FALSE, row.names = FALSE)
  return(points)
} 

# Random initialization of a solution
# random coordinates for n squares between (0,0) and (1000-c, 1000-c)
init <- function() {
  return(matrix(ceiling(runif(n=2 * n, min = 0, max = 1000 - c)), ncol = 2, byrow = TRUE))  
}

# create a mutation operator:
# p: mutation rate of each square
# r: range of coordinates mutation
mutation_create <- function(p, r) {
  values <- c(-r:r)
  mut <- function(x) {
    y <- x
    for(i in 1:nrow(x)) {
      if (runif(n=1) < p) {
        delta <- sample(values, size = 2, replace = TRUE)
        while (delta[1] == 0 && delta[2] == 0)
          delta <- sample(values, size = 2, replace = TRUE)
        y[i,] <- y[i,] + delta
      }
    }
    return(y)
  }
  
  return(mut)
}

# Perform a random walk and return fitness values
random_walk_2 <- function(fitness, mutation, wlen) {
  x <- init()
  # vector of fitness values of the random walk
  fit_values <- matrix(fitness(x), ncol = 2)
  
  for(i in 1:(wlen-1)) {
    x <- mutation(x)
    fit_values <- rbind(fit_values, fitness(x))
  }
  
  return(fit_values)
}

# Pareto Hill-Climbing
pareto_hc <- function(fitness, mutation, size, x) {
  i = 1
  
  # fitness value of the current solution x
  fitx = fitness(x)
  
  step    = 0  # number of step of the adaptive walk
  neval = 0  # number of evaluation
  while (i <= size) {
    # one random neighbor y
    y <- mutation(x)
    fity <- fitness(y)
    neval <- neval + 1 
    if (fitx[1] < fity[1] && fitx[2] < fity[2]) {
      # accept the neighbor y
      x <- y
      fitx = fity
      step <- step + 1
      #      cat(step, i, fit, "\n")
      # start again
      i <- 1
    } else {
      # next neighbor
      i <- i + 1  
    }
  }
  
  # return the local optimum with the statistics
  return(list(x = x, fitness = fitx, step = step, neval = neval))
}

# Pareto adaptive walks
pareto_adaptive_length <- function(fitness, mutation, nb) {
  # data frame for the results
  results <- data.frame(step=integer(), neval=integer(), f1=integer(), f2=integer())
  
  for(i in 1:nb) {
    # initial random solution
    x <- init()
    # run hc
    result <- pareto_hc(fitness, mutation, 200, x)
    # print and save the result
    cat("step:", result$step, "neval:", result$neval, "fitness:", result$fitness, "x:", result$x, "\n")
    results <- rbind(results, data.frame(step=result$step, neval=result$neval, f1=result$fitness[1], f2=result$fitness[2]))
  }
  
  return(results)
}

# Perform a Pareto adaptive walk on each SP2 problem
phc_samplings <- function() {
  set.seed(12345)
  mutation <- mutation_create(p = 0.2, r = 40)
  
  for(i in 11:20) {
    cat("pb", i, "\n")
    # fitness function of the SP problem i
    points <- read.table(paste0("data/pb_", i,".csv"), sep = " ", header = TRUE)
    ff <- fitness2_create(points)
    
    # random walk
    results <- pareto_adaptive_length(ff, mutation, 30)
    
    # save into file
    write.table(results, file = paste0("data/sp2_phc_02_40_", i, ".csv"), quote = FALSE, row.names = FALSE)
  }
}

# Compute the neutral rate for a biobjective problem estimated by a random walk
neutral_rate <- function(fit_values) {
  l <- nrow(fit_values)
  
  nrate = sum(fit_values[1:l-1,1] == fit_values[2:l,1] && fit_values[1:l-1,2] == fit_values[2:l,2]) / (l-1)
  
  return(nrate)
}

hypervolume_solution <- function(fvalue) {
  return(fvalue[1] * fvalue[2])  
}

# Compute main fitness landscape features based on a random random
fitness_landscape2_features <- function(fit_values) {
  l <- nrow(fit_values)
  
  hv <- apply(fit_values, 1, hypervolume_solution)
  
  # autocorrelation function
  acf1 <- acf(hv, plot = FALSE, lag.max = 400)
  # autocorrelation length
  #len1 <- which.max(acf1$acf < 4 / sqrt(l)) - 1
  #if (len1 == 0) len1 = 400
  
  # all statistics
  rho = cor(fit_values[,1], fit_values[,2])
  
  res <- list(avg1 = mean(fit_values[,1]), avg2 = mean(fit_values[,2]), sd1 = sd(fit_values[,1]), sd2 = sd(fit_values[,2]), rho = rho, rho1 = acf1$acf[2], nrate = neutral_rate(fit_values))
  
  return(res)
}

# Perform a random walk (of length 1000) on each SP problem
random_walk2_samplings <- function() {
  set.seed(12345)
  mutation <- mutation_create(p = 0.2, r = 40)
  
  for(i in 1:20) {
    cat("pb", i, "\n")
    # fitness function of the SP problem i
    points <- read.table(paste0("data/pb_", i,".csv"), sep = " ", header = TRUE)
    ff <- fitness2_create(points)
    
    # random walk
    results <- random_walk_2(ff, mutation, 1000)
    
    # save into file
    write.table(results, file = paste0("data/sp2_rw_02_40_", i, ".csv"), quote = FALSE, row.names = FALSE, col.names = FALSE)
  }
}
  
fitness_landscape2_analysis <- function() {
  # analysis of each SP2 problem
  df <- data.frame()
  for(i in 1:20) {
    values = read.table(paste0("data/sp2_rw_02_40_", i,".csv"), sep = " ", header = FALSE)
    res <- fitness_landscape2_features(values)
    res$pb = i
    df <- rbind(df, res)
  }
    
  # Pareto adaptive length
  v <- c()
  for(i in 1:20) {
    values = read.table(paste0("data/sp2_phc_02_40_", i,".csv"), sep = " ", header = TRUE)
    res <- mean(values$step)
    v <- rbind(v, res)
  }
  df$alength = v 
  
  return(df)
}

main <- function() {
  # create points from random or from a file SP1
  points1 <- pb1_create()
  #points1 <- read.table("pb_1.csv", sep = " ", header = TRUE)
  p <- draw(points1) 
  bff1 <- fitness2_create(points1)
  
  # create points from random or from a file SP2
  points2 <- pb2_create() 
  #points2 <- read.table("pb_2.csv", sep = " ", header = TRUE)
  p <- draw(points2) 
  bff2 <- fitness2_create(points2)
  
  # easy test
  x <- init()
  cat(bff1(x), "\n")
  cat(bff2(x), "\n")
  
  # random walk on each problem
  #random_walk2_samplings()
  
  # Pareto Hill-climber
  #phc_samplings()
  
  # fitness landscape analysis
  df <- fitness_landscape2_analysis()
  
  # repair NA values
  df[5,]$rho <- 0
  df[5,]$rho1 <- 0
  
  # Analyze the results
  View(df)
  ggpairs(df)
  
}

